--[[
Camera utility module

grabbed from http://nova-fusion.com/2011/04/19/cameras-in-love2d-part-1-the-basics/

used to make an easy camera system

Use example:
============================================================
require "camera"

local cam = Camera()

function love.draw()
	cam:set()
	-- all drawings under this camera go here
	cam:unset()
end
============================================================
]]

local function clamp (val,min,max)
	return val < min and min or (val > max and max or val)
end

Camera = {}
Camera.__index = Camera

--[[
Creates a new camera
_x  = camera X pos
_y  = camera Y pos
_sx = camera scale in X
_sy = camera scale in Y
_r  = camera rotation degrees
]]
function Camera.new(_x,_y,_sx,_sy,_r)
	local c = {
		_x = _x or 0,
		_y = _y or 0,
		scaleX = _sx or 1,
		scaleY = _sy or 1,
		rotation = _r or 0,
		_layers = {},
	}
	setmetatable(c,Camera)
	return c
end
setmetatable(Camera, {__call = function(_,...) return Camera.new(...) end})

function Camera:set()
	love.graphics.push()
	love.graphics.rotate(-self.rotation)
	love.graphics.scale(1 / self.scaleX, 1 / self.scaleY)
	love.graphics.translate(-self._x, -self._y)
end

function Camera:unset()
	love.graphics.pop()
end

function Camera:draw()
	local bx, by = self._x, self._y
	for _,v in ipairs(self._layers) do
		self._x = math.floor(bx*v.scale)
		self._y = math.floor(by*v.scale)
		self:set()
		v.draw()
		self:unset()
	end
end

function Camera:setX(x)
	if self._bounds then
		self._x = math.floor(clamp(x, self._bounds.x1, self._bounds.x2))
	else
		self._x = math.floor(x)
	end
end

function Camera:setY(y)
	if self._bounds then
		self._y = math.floor(clamp(y, self._bounds.y1, self._bounds.y2))
	else
		self._y = math.floor(y)
	end
end

function Camera:getX() return self._x end
function Camera:getY() return self._y end

-- to support screen split
function Camera:getHeight()
	return love.graphics.getHeight()
end

function Camera:getWidth()
	return love.graphics.getWidth()
end

function Camera:move(dx, dy)
	self:setX(self._x + (dx or 0))
	self:setY(self._y + (dy or 0))
end

function Camera:rotate(dr)
	self.rotation = self.rotation + (dr or 0)
end

function Camera:scale(sx, sy)
	sx = sx or 1
	self.scaleX = self.scaleX * sx
	self.scaleY = self.scaleY * (sy or sx)
end

function Camera:setPosition(x, y)
	self:setX(x)
	self:setY(y)
end

function Camera:setScale(sx, sy)
	self.scaleX = sx or self.scaleX
	self.scaleY = sy or self.scaleY
end

function Camera:setBounds(x1, y1, x2, y2)
	self._bounds = {x1 = x1, 
					y1 = y1, 
					x2 = x2, 
					y2 = y2}
end

function Camera:getBounds() 
	return unpack(self._bounds) 
end

-- parallax utility
function Camera:newLayer(scale, func)
	table.insert(self._layers,{draw = func, scale = scale})
	table.sort(self._layers, function(a,b) return a.scale < b.scale end)
end

-- used to get camera-relative mouse position
function Camera:mousePosition()
	return love.mouse.getX() * self.scaleX + self._x, love.mouse.getY() * self.scaleY + self._y
end