--[[
Tilemap module

Used to map a scenario using a sprite atlas and a tile map
A tilemap accepts an arbitrary number of map layers.

Use example:
============================================================
require "tilemap"
local map = {
{0,0,3,0,0},
{0,2,4,2,0},
{1,1,1,1,1}
}

local tilemap
function love.load()
	tilemap = Tilemap.new("tilemap.png",16,16,0,0,128,128)
	tilemap:AddMap(map)
end

function love.draw()
	tilemap:draw()
end
============================================================

collision map: holds collision info for each tile in atlas

value | blocks from
------+------------
1     | upper
2     | left
4     | right
8     | bottom
0x11  | /|
0x12  | |\
0x14  | \|
0x18  | |/
]]

local function hex(number)
	return tonumber(""..number,16)
end

-- scale conversion
local function scale(sf, sc, s, df, dc)
	-- (s-sf)/(sc-sf) = (ret - df)/(dc-df)
	-- ret - df = (s-sf)(dc-df)/(sc-sf)
	-- ret = (s-sf)(dc-df)/(sc-sf) + df
	return (s-sf)*(dc-df)/(sc-sf) + df
end

Tilemap = {}
Tilemap.__index = Tilemap
--[[ 
Creates a new tileset

tileset   = tileset atlas
tileSizeX = single tile X size
tileSizeY = single tile Y size
left      = tilemap x offset
top       = tilemap y offset
width     = tilemap render width
height    = tilemap render height
]]
function Tilemap.new(tileset, tileSizeX, tileSizeY, left, top)
	-- private vars
	local _maps = {} -- tile map
	local _atlas = love.graphics.newImage(tileset) -- tiles atlas
	local _quads = {}
	local _collisionMap = {}
	-- public vars and methods
	local t = {
		x = top or 0, -- draw offset
		y = left or 0,
		
		-- add map to tilemap list (lowest order is back, highest is in front)
		AddMap = function(self,mapTable, order)
			if order ~= nil and order > 0 then
				table.insert(_maps,order,mapTable)
			else
				table.insert(_maps,mapTable)
			end
		end,
		
		SetCollisionMap = function(self,map)
			_collisionMap = map
		end,
		
		GetNextCollision = function(self,x,y,dir)
			local xTile = math.floor(x / tileSizeX) +1
			local yTile = math.floor(y / tileSizeY) +1
			local map = _maps[1]
			local lookup = {
				left = function()
					if xTile == 1 then return 0 end
					xTile = xTile - 1
					while xTile > 0 do
						if _collisionMap[map[yTile][xTile]] then
							if bit.band(_collisionMap[map[yTile][xTile]],hex(10)) ~= 0 then -- diagonal
								local diag = bit.band(_collisionMap[map[yTile][xTile]],hex("0F"))
								if bit.band(diag,hex(2)) ~= 0 then -- |\
									return math.floor(scale((yTile - 1)*tileSizeY, (yTile)*tileSizeY, y, (xTile - 1)*tileSizeX, (xTile)*tileSizeX)) -- give left collision, given current height
								elseif bit.band(diag,hex(8)) ~= 0 then -- |/
									return math.floor(scale((yTile - 1)*tileSizeY, (yTile)*tileSizeY, y, (xTile)*tileSizeX, (xTile - 1)*tileSizeX))
								end
							elseif bit.band(_collisionMap[map[yTile][xTile]],hex(2)) ~= 0 then -- right collision
								return (xTile)*tileSizeX
							end
						end
						xTile = xTile - 1
					end
					return 0
				end,
			
				right = function()
					if xTile == self.width/tileSizeX then return self.width end
					xTile = xTile + 1
					while xTile <= #map[yTile] do
						if _collisionMap[map[yTile][xTile]] then
							if bit.band(_collisionMap[map[yTile][xTile]],hex(10)) ~= 0 then -- diagonal
								local diag = bit.band(_collisionMap[map[yTile][xTile]],hex("0F"))
								if bit.band(diag,hex(1)) ~= 0 then -- /|
									return math.floor(scale((yTile - 1)*tileSizeY, (yTile)*tileSizeY, y, (xTile)*tileSizeX, (xTile - 1)*tileSizeX)) -- give left collision, given current height
								elseif bit.band(diag,hex(4)) ~= 0 then -- \|
									return math.floor(scale((yTile - 1)*tileSizeY, (yTile)*tileSizeY, y, (xTile - 1)*tileSizeX, (xTile)*tileSizeX))
								end
							elseif bit.band(_collisionMap[map[yTile][xTile]],4) ~= 0 then -- left collision
								return (xTile-1)*tileSizeX
							end
						end
						xTile = xTile + 1
					end
					return self.width
				end,
			
				up = function()
					if yTile == 1 then return 0 end
					yTile = yTile - 1
					while yTile > 0 do
						if _collisionMap[map[yTile][xTile]] then
							if bit.band(_collisionMap[map[yTile][xTile]],hex(10)) ~= 0 then -- diagonal
								local diag = bit.band(_collisionMap[map[yTile][xTile]],hex("0F"))
								if bit.band(diag,hex(4)) ~= 0 then -- \|
									return math.floor(scale((xTile - 1)*tileSizeX, (xTile)*tileSizeX, x, (yTile - 1)*tileSizeY, (yTile)*tileSizeY)) -- give left collision, given current height
								elseif bit.band(diag,hex(8)) ~= 0 then -- |/
									return math.floor(scale((xTile - 1)*tileSizeX, (xTile)*tileSizeX, x, (yTile)*tileSizeY, (yTile - 1)*tileSizeY))
								end
							elseif bit.band(_collisionMap[map[yTile][xTile]],1) ~= 0 then -- down collision
								return yTile*tileSizeY
							end
						end
						yTile = yTile - 1
					end
					return 0
				end,
			
				down = function()
					yTile = yTile + 1
					while yTile <= #map do
						if _collisionMap[map[yTile][xTile]] then
							if bit.band(_collisionMap[map[yTile][xTile] ],hex(10)) ~= 0 then -- diagonal
								local diag = bit.band(_collisionMap[map[yTile][xTile] ],hex("0F"))
								if bit.band(diag,hex(1)) ~= 0 then -- /|
									return math.floor(scale((xTile - 1)*tileSizeX, (xTile)*tileSizeX, x, (yTile)*tileSizeY, (yTile - 1)*tileSizeY))
								elseif bit.band(diag,hex(2)) ~= 0 then -- |\
									return math.floor(scale((xTile - 1)*tileSizeX, (xTile)*tileSizeX, x, (yTile - 1)*tileSizeY, (yTile)*tileSizeY))
								end
							elseif bit.band(_collisionMap[map[yTile][xTile]],8) ~= 0 then -- up collision
								return (yTile-1)*tileSizeY
							end
						end
						yTile = yTile + 1
					end
					return self.height
				end,
			}
			if lookup[dir] then return lookup[dir]() end
			return nil
		end,

		-- remove tilemap from tiles list
		RemoveMap = function(self,order)
			table.remove(_maps,order)
		end,
		
		-- draw tilemap on screen
		draw = function(self)
			for _,map in ipairs(_maps) do
				for y,v in ipairs(map) do
					for x,m in ipairs(v) do
						-- draw only if there is something to draw
						if _quads[m] then
							-- and only if the tile will be visible
							local tilePosX = (x-1)*tileSizeX + self.x
							local tilePosY = (y-1)*tileSizeY + self.y
								love.graphics.drawq(_atlas, 
													_quads[m],
													math.floor(tilePosX),
													math.floor(tilePosY))
						end
					end
				end
			end
		end,
	}
	local tileCountX = _atlas:getWidth()/tileSizeX
	local tileCountY = _atlas:getHeight()/tileSizeY
	
	for y=1,tileCountY do
		for x=1,tileCountX do
			table.insert(_quads,love.graphics.newQuad((x-1)*tileSizeX,
														(y-1)*tileSizeY,
														tileSizeX,
														tileSizeY,
														_atlas:getWidth(),
														_atlas:getHeight()))
		end
	end
	
	setmetatable(t,{
		__index = function(t,key)
			--lookup table
			local cases = {
				height = function()
					local maxH = 0
					for k,m in ipairs(_maps) do
						if #m > maxH then maxH = #m end
					end
					return maxH * tileSizeY
				end,
				width = function()
					local maxW = 0
					for k,m in ipairs(_maps) do
						for _,v in ipairs(m) do
							if #v > maxW then maxW = #v end
						end
					end
					return maxW * tileSizeX
				end,
			}
			if cases[key] then return cases[key]()
			else return rawget(t,key) end
		end
	})
	return t
end
setmetatable(Tilemap, {__call = function(_,...) return Tilemap.new(...) end})
